TEST_TEMP_RAW = 529191
TEST_TEMP_CMP = 24.7894877676

TEST_PRES_RAW = 326816
TEST_PRES_CMP = 1006.61517564
TEST_ALT_CMP = 57.3174


def test_temperature():
    from calibration import BMP280Calibration
    from tools import SMBusFakeDevice

    from bmp280 import BMP280
    dev = SMBusFakeDevice(1)

    # Load the fake temperature into the virtual registers
    dev.regs[0xfc] = (TEST_TEMP_RAW & 0x0000F) << 4
    dev.regs[0xfb] = (TEST_TEMP_RAW & 0x00FF0) >> 4
    dev.regs[0xfa] = (TEST_TEMP_RAW & 0xFF000) >> 12

    bmp280 = BMP280(i2c_dev=dev)
    bmp280.setup()

    # Replace the loaded calibration with our known values
    bmp280.calibration = BMP280Calibration()

    assert round(bmp280.get_temperature(), 4) == round(TEST_TEMP_CMP, 4)


def test_temperature_forced():
    from calibration import BMP280Calibration
    from tools import SMBusFakeDevice

    from bmp280 import BMP280
    dev = SMBusFakeDevice(1)

    # Load the fake temperature into the virtual registers
    dev.regs[0xfc] = (TEST_TEMP_RAW & 0x0000F) << 4
    dev.regs[0xfb] = (TEST_TEMP_RAW & 0x00FF0) >> 4
    dev.regs[0xfa] = (TEST_TEMP_RAW & 0xFF000) >> 12

    bmp280 = BMP280(i2c_dev=dev)
    bmp280.setup(mode="forced")

    # Replace the loaded calibration with our known values
    bmp280.calibration = BMP280Calibration()

    assert round(bmp280.get_temperature(), 4) == round(TEST_TEMP_CMP, 4)


def test_pressure():
    from calibration import BMP280Calibration
    from tools import SMBusFakeDevice

    from bmp280 import BMP280
    dev = SMBusFakeDevice(1)

    # Load the fake temperature values into the virtual registers
    # Pressure is temperature compensated!!!
    dev.regs[0xfc] = (TEST_TEMP_RAW & 0x0000F) << 4
    dev.regs[0xfb] = (TEST_TEMP_RAW & 0x00FF0) >> 4
    dev.regs[0xfa] = (TEST_TEMP_RAW & 0xFF000) >> 12

    # Load the fake pressure values
    dev.regs[0xf9] = (TEST_PRES_RAW & 0x0000F) << 4
    dev.regs[0xf8] = (TEST_PRES_RAW & 0x00FF0) >> 4
    dev.regs[0xf7] = (TEST_PRES_RAW & 0xFF000) >> 12

    bmp280 = BMP280(i2c_dev=dev)
    bmp280.setup()

    # Replace the loaded calibration with our known values
    bmp280.calibration = BMP280Calibration()

    assert round(bmp280.get_pressure(), 4) == round(TEST_PRES_CMP, 4)


def test_altitude():
    from calibration import BMP280Calibration
    from tools import SMBusFakeDevice

    from bmp280 import BMP280
    dev = SMBusFakeDevice(1)

    # Load the fake temperature values into the virtual registers
    # Pressure is temperature compensated!!!
    dev.regs[0xfc] = (TEST_TEMP_RAW & 0x0000F) << 4
    dev.regs[0xfb] = (TEST_TEMP_RAW & 0x00FF0) >> 4
    dev.regs[0xfa] = (TEST_TEMP_RAW & 0xFF000) >> 12

    # Load the fake pressure values
    dev.regs[0xf9] = (TEST_PRES_RAW & 0x0000F) << 4
    dev.regs[0xf8] = (TEST_PRES_RAW & 0x00FF0) >> 4
    dev.regs[0xf7] = (TEST_PRES_RAW & 0xFF000) >> 12

    bmp280 = BMP280(i2c_dev=dev)
    bmp280.setup()

    # Replace the loaded calibration with our known values
    bmp280.calibration = BMP280Calibration()

    assert round(bmp280.get_altitude(), 4) == round(TEST_ALT_CMP, 4)



"""BMP280 Driver."""

import struct
import time

from i2cdevice import BitField, Device, Register, _int_to_bytes
from i2cdevice.adapter import Adapter, LookupAdapter

__version__ = "1.0.0"

CHIP_ID = 0x58
I2C_ADDRESS_GND = 0x76
I2C_ADDRESS_VCC = 0x77


class S16Adapter(Adapter):
    """Convert unsigned 16bit integer to signed."""

    def _decode(self, value):
        return struct.unpack("<h", _int_to_bytes(value, 2))[0]


class U16Adapter(Adapter):
    """Convert from bytes to an unsigned 16bit integer."""

    def _decode(self, value):
        return struct.unpack("<H", _int_to_bytes(value, 2))[0]


class BMP280Calibration:
    def __init__(self):
        self.dig_t1 = 0
        self.dig_t2 = 0
        self.dig_t3 = 0

        self.dig_p1 = 0
        self.dig_p2 = 0
        self.dig_p3 = 0
        self.dig_p4 = 0
        self.dig_p5 = 0
        self.dig_p6 = 0
        self.dig_p7 = 0
        self.dig_p8 = 0
        self.dig_p9 = 0

        self.temperature_fine = 0

    def set_from_namedtuple(self, value):
        # Iterate through a tuple supplied by i2cdevice
        # and copy its values into the class attributes
        for key in self.__dict__.keys():
            try:
                setattr(self, key, getattr(value, key))
            except AttributeError:
                pass

    def compensate_temperature(self, raw_temperature):
        var1 = (raw_temperature / 16384.0 - self.dig_t1 / 1024.0) * self.dig_t2
        var2 = raw_temperature / 131072.0 - self.dig_t1 / 8192.0
        var2 = var2 * var2 * self.dig_t3
        self.temperature_fine = var1 + var2
        return self.temperature_fine / 5120.0

    def compensate_pressure(self, raw_pressure):
        var1 = self.temperature_fine / 2.0 - 64000.0
        var2 = var1 * var1 * self.dig_p6 / 32768.0
        var2 = var2 + var1 * self.dig_p5 * 2
        var2 = var2 / 4.0 + self.dig_p4 * 65536.0
        var1 = (self.dig_p3 * var1 * var1 / 524288.0 + self.dig_p2 * var1) / 524288.0
        var1 = (1.0 + var1 / 32768.0) * self.dig_p1
        pressure = 1048576.0 - raw_pressure
        pressure = (pressure - var2 / 4096.0) * 6250.0 / var1
        var1 = self.dig_p9 * pressure * pressure / 2147483648.0
        var2 = pressure * self.dig_p8 / 32768.0
        return pressure + (var1 + var2 + self.dig_p7) / 16.0


class BMP280:
    def __init__(self, i2c_addr=I2C_ADDRESS_GND, i2c_dev=None):
        self.calibration = BMP280Calibration()
        self._is_setup = False
        self._i2c_addr = i2c_addr
        self._i2c_dev = i2c_dev
        self._bmp280 = Device([I2C_ADDRESS_GND, I2C_ADDRESS_VCC], i2c_dev=self._i2c_dev, bit_width=8, registers=(
            Register("CHIP_ID", 0xD0, fields=(
                BitField("id", 0xFF),
            )),
            Register("RESET", 0xE0, fields=(
                BitField("reset", 0xFF),
            )),
            Register("STATUS", 0xF3, fields=(
                BitField("measuring", 0b00001000),  # 1 when conversion is running
                BitField("im_update", 0b00000001),  # 1 when NVM data is being copied
            )),
            Register("CTRL_MEAS", 0xF4, fields=(
                BitField("osrs_t", 0b11100000,   # Temperature oversampling
                         adapter=LookupAdapter({
                             1: 0b001,
                             2: 0b010,
                             4: 0b011,
                             8: 0b100,
                             16: 0b101
                         })),
                BitField("osrs_p", 0b00011100,   # Pressure oversampling
                         adapter=LookupAdapter({
                             1: 0b001,
                             2: 0b010,
                             4: 0b011,
                             8: 0b100,
                             16: 0b101})),
                BitField("mode", 0b00000011,     # Power mode
                         adapter=LookupAdapter({
                             "sleep": 0b00,
                             "forced": 0b10,
                             "normal": 0b11})),
            )),
            Register("CONFIG", 0xF5, fields=(
                BitField("t_sb", 0b11100000,     # Temp standby duration in normal mode
                         adapter=LookupAdapter({
                             0.5: 0b000,
                             62.5: 0b001,
                             125: 0b010,
                             250: 0b011,
                             500: 0b100,
                             1000: 0b101,
                             2000: 0b110,
                             4000: 0b111})),
                BitField("filter", 0b00011100),                   # Controls the time constant of the IIR filter
                BitField("spi3w_en", 0b0000001, read_only=True),  # Enable 3-wire SPI interface when set to 1. IE: Don't set this bit!
            )),
            Register("DATA", 0xF7, fields=(
                BitField("temperature", 0x000000FFFFF0),
                BitField("pressure", 0xFFFFF0000000),
            ), bit_width=48),
            Register("CALIBRATION", 0x88, fields=(
                BitField("dig_t1", 0xFFFF << 16 * 11, adapter=U16Adapter()),   # 0x88 0x89
                BitField("dig_t2", 0xFFFF << 16 * 10, adapter=S16Adapter()),   # 0x8A 0x8B
                BitField("dig_t3", 0xFFFF << 16 * 9, adapter=S16Adapter()),    # 0x8C 0x8D
                BitField("dig_p1", 0xFFFF << 16 * 8, adapter=U16Adapter()),    # 0x8E 0x8F
                BitField("dig_p2", 0xFFFF << 16 * 7, adapter=S16Adapter()),    # 0x90 0x91
                BitField("dig_p3", 0xFFFF << 16 * 6, adapter=S16Adapter()),    # 0x92 0x93
                BitField("dig_p4", 0xFFFF << 16 * 5, adapter=S16Adapter()),    # 0x94 0x95
                BitField("dig_p5", 0xFFFF << 16 * 4, adapter=S16Adapter()),    # 0x96 0x97
                BitField("dig_p6", 0xFFFF << 16 * 3, adapter=S16Adapter()),    # 0x98 0x99
                BitField("dig_p7", 0xFFFF << 16 * 2, adapter=S16Adapter()),    # 0x9A 0x9B
                BitField("dig_p8", 0xFFFF << 16 * 1, adapter=S16Adapter()),    # 0x9C 0x9D
                BitField("dig_p9", 0xFFFF << 16 * 0, adapter=S16Adapter()),    # 0x9E 0x9F
            ), bit_width=192)
        ))

    def setup(self, mode="normal", temperature_oversampling=16, pressure_oversampling=16, temperature_standby=500):
        if self._is_setup:
            return
        self._is_setup = True

        self._bmp280.select_address(self._i2c_addr)
        self._mode = mode

        if mode == "forced":
            mode = "sleep"

        try:
            chip = self._bmp280.get("CHIP_ID")
            if chip.id != CHIP_ID:
                raise RuntimeError(f"Unable to find bmp280 on 0x{self._i2c_addr:02x}, CHIP_ID returned {chip.id:02x}")
        except IOError:
            raise RuntimeError(f"Unable to find bmp280 on 0x{self._i2c_addr:02x}, IOError")

        self._bmp280.set("CTRL_MEAS",
                         mode=mode,
                         osrs_t=temperature_oversampling,
                         osrs_p=pressure_oversampling)

        self._bmp280.set("CONFIG",
                         t_sb=temperature_standby,
                         filter=2)

        self.calibration.set_from_namedtuple(self._bmp280.get("CALIBRATION"))

    def update_sensor(self):
        self.setup()

        if self._mode == "forced":
            # Trigger a reading in forced mode and wait for result
            self._bmp280.set("CTRL_MEAS", mode="forced")
            while self._bmp280.get("STATUS").measuring:
                time.sleep(0.001)

        raw = self._bmp280.get("DATA")

        self.temperature = self.calibration.compensate_temperature(raw.temperature)
        self.pressure = self.calibration.compensate_pressure(raw.pressure) / 100.0

    def get_temperature(self):
        self.update_sensor()
        return self.temperature

    def get_pressure(self):
        self.update_sensor()
        return self.pressure

    def get_altitude(self, qnh=1013.25, manual_temperature=None):
        # qnh = pressure at sea level where the readings are being taken.
        # The temperature should be the outdoor temperature.
        # Use the manual_temperature variable if temperature adjustments are required.
        self.update_sensor()
        pressure = self.get_pressure()
        if manual_temperature is None:
            temperature = self.get_temperature()
        else:
            temperature = manual_temperature
        altitude = ((pow((qnh / pressure), (1.0 / 5.257)) - 1) * (temperature + 273.15)) / 0.0065
        return altitude



import time

from smbus2 import SMBus

from bmp280 import BMP280

bus = SMBus(1)
bmp280 = BMP280(i2c_dev=bus)

# Set up in "forced" mode
# In this mode `get_temperature` and `get_pressure` will trigger
# a new reading and wait for the result.
# The chip will return to sleep mode when finished.
bmp280.setup(mode="forced")

while True:
    temperature = bmp280.get_temperature()
    print("{temperature:05.2f}*C")
    time.sleep(1)



import bmp280


class BMP280Calibration(bmp280.BMP280Calibration):
    """Prefil the calibration class with known values."""
    def __init__(self):
        bmp280.BMP280Calibration.__init__(self)

        self.dig_t1 = 28009
        self.dig_t2 = 25654
        self.dig_t3 = 50

        self.dig_p1 = 39145
        self.dig_p2 = -10750
        self.dig_p3 = 3024
        self.dig_p4 = 5667
        self.dig_p5 = -120
        self.dig_p6 = -7
        self.dig_p7 = 15500
        self.dig_p8 = -14600
        self.dig_p9 = 6000



#!/usr/bin/env python

import time

from smbus2 import SMBus

from bmp280 import BMP280

print("""temperature-and-pressure.py - Displays the temperature and pressure.

Press Ctrl+C to exit!

""")

# Initialise the BMP280
bus = SMBus(1)
bmp280 = BMP280(i2c_dev=bus)

while True:
    temperature = bmp280.get_temperature()
    pressure = bmp280.get_pressure()
    print(f"{temperature:05.2f}*C {pressure:05.2f}hPa")
    time.sleep(1)



#!/usr/bin/env python

import time
from subprocess import PIPE, Popen

from smbus2 import SMBus

from bmp280 import BMP280

print("""compensated-temperature.py - Use the CPU temperature to compensate temperature
readings from the BMP280 sensor. Method adapted from Initial State's Enviro pHAT
review: https://medium.com/@InitialState/tutorial-review-enviro-phat-for-raspberry-pi-4cd6d8c63441

Press Ctrl+C to exit!

""")

# Initialise the BMP280
bus = SMBus(1)
bmp280 = BMP280(i2c_dev=bus)


# Gets the CPU temperature in degrees C
def get_cpu_temperature():
    process = Popen(["vcgencmd", "measure_temp"], stdout=PIPE)
    output, _error = process.communicate()
    return float(output[output.index("=") + 1 : output.rindex("'")])


factor = 1.2  # Smaller numbers adjust temp down, vice versa
smooth_size = 10  # Dampens jitter due to rapid CPU temp changes

cpu_temps = []

while True:
    cpu_temp = get_cpu_temperature()
    cpu_temps.append(cpu_temp)

    if len(cpu_temps) > smooth_size:
        cpu_temps = cpu_temps[1:]

    smoothed_cpu_temp = sum(cpu_temps) / float(len(cpu_temps))
    raw_temp = bmp280.get_temperature()
    comp_temp = raw_temp - ((smoothed_cpu_temp - raw_temp) / factor)

    print(f"Compensated temperature: {comp_temp:05.2f} *C")

    time.sleep(1.0)



import sys

import mock
import pytest


def test_setup_not_present():
    sys.modules['smbus2'] = mock.MagicMock()
    from bmp280 import BMP280
    bmp280 = BMP280()
    with pytest.raises(RuntimeError):
        bmp280.setup()


def test_setup_mock_present():
    from tools import SMBusFakeDevice
    smbus = mock.Mock()
    smbus.SMBus = SMBusFakeDevice
    sys.modules['smbus2'] = smbus
    from bmp280 import BMP280
    bmp280 = BMP280()
    bmp280.setup()


def test_setup_forced_mode():
    from tools import SMBusFakeDevice
    smbus = mock.Mock()
    smbus.SMBus = SMBusFakeDevice
    sys.modules['smbus2'] = smbus
    from bmp280 import BMP280
    bmp280 = BMP280()
    bmp280.setup(mode="forced")



#!/usr/bin/env python

from smbus2 import SMBus

from bmp280 import BMP280

print("""dump-calibration.py - Dumps calibration data.

Press Ctrl+C to exit!

""")

# Initialise the BMP280
bmp280 = BMP280(i2c_dev=SMBus(1))
bmp280.setup()

for key in dir(bmp280.calibration):
    if key.startswith("dig_"):
        value = getattr(bmp280.calibration, key)
        print(f"{key} = {value}")



#!/usr/bin/env python

import time

from smbus2 import SMBus

from bmp280 import BMP280

print("""relative-altitude.py - Calculates relative altitude from pressure.

Press Ctrl+C to exit!

""")

# Initialise the BMP280
bus = SMBus(1)
bmp280 = BMP280(i2c_dev=bus)

baseline_values = []
baseline_size = 100

print(f"Collecting baseline values for {baseline_size:d} seconds. Do not move the sensor!\n")

for i in range(baseline_size):
    pressure = bmp280.get_pressure()
    baseline_values.append(pressure)
    time.sleep(1)

baseline = sum(baseline_values[:-25]) / len(baseline_values[:-25])

while True:
    altitude = bmp280.get_altitude(qnh=baseline)
    print(f"Relative altitude: {altitude:05.2f} metres")
    time.sleep(1)



import time

from bmp280 import CHIP_ID, _bmp280

print('0x{:02x} {}'.format(_bmp280.CHIP_ID.get_id(), 'ok' if _bmp280.CHIP_ID.get_id() == CHIP_ID else 'err'))

with _bmp280.CTRL_MEAS as reg:
    reg.set_mode('normal')
    reg.set_osrs_t(16)  # 16x temperature oversampling
    reg.set_osrs_p(16)  # 16x pressure oversampling
    reg.write()

with _bmp280.CONFIG as reg:
    reg.set_t_sb(500)
    reg.set_filter(2)
    reg.write()

with _bmp280.CALIBRATION as reg:
    dig_t1 = reg.get_dig_t1()
    dig_t2 = reg.get_dig_t2()
    dig_t3 = reg.get_dig_t3()

    dig_p1 = reg.get_dig_p1()
    dig_p2 = reg.get_dig_p2()
    dig_p3 = reg.get_dig_p3()
    dig_p4 = reg.get_dig_p4()
    dig_p5 = reg.get_dig_p5()
    dig_p6 = reg.get_dig_p6()
    dig_p7 = reg.get_dig_p7()
    dig_p8 = reg.get_dig_p8()
    dig_p9 = reg.get_dig_p9()

    print(dig_t1, dig_t2, dig_t3)
    print(dig_p1, dig_p2, dig_p3, dig_p4, dig_p5, dig_p6, dig_p7, dig_p8, dig_p9)

try:
    while True:
        raw_temp = _bmp280.DATA.get_temperature()
        raw_pres = _bmp280.DATA.get_pressure()

        print('0b{:020b}'.format(raw_temp))

        var1 = (raw_temp / 16384.0 - dig_t1 / 1024.0) * dig_t2
        var2 = (raw_temp / 131072.0 - dig_t1 / 8192.0) * (raw_temp / 131072.0 - dig_t1 / 8192.0) * dig_t3
        temp = (var1 + var2) / 5120.0
        t_fine = (var1 + var2)

        print(temp)
        time.sleep(1)

except KeyboardInterrupt:
    pass
